import logging.config
from logging import Logger
from os import path

log_file_path = path.join(path.dirname(path.abspath(__file__)), "resources/logger.conf")
logging.config.fileConfig(log_file_path)


def get_logger() -> Logger:
    return logging.getLogger()



import itertools
from typing import Any, Optional

import matplotlib.patches as patches
import matplotlib.pyplot as plt
import numpy as np
from mpl_toolkits.axes_grid1 import make_axes_locatable

from receptivefield.common import estimate_rf_from_gradient
from receptivefield.image import get_default_image
from receptivefield.types import (
    ImageShape,
    GridPoint,
    GridShape,
    ReceptiveFieldDescription,
    ReceptiveFieldRect,
    to_rf_rect,
)


def _plot_rect(
    ax,
    rect: ReceptiveFieldRect,
    color: Any,
    alpha: float = 0.9,
    linewidth: float = 5,
    size: float = 90,
) -> None:
    """
    Plot rectangle and center point.

    :param ax: matplotlib axis
    :param rect: definition of rectangle
    :param color:
    :param alpha:
    :param linewidth:
    :param size: point size
    """
    ax.add_patch(
        patches.Rectangle(
            (rect.y - rect.h / 2, rect.x - rect.w / 2),
            rect.h,
            rect.w,
            alpha=alpha,
            fill=False,
            facecolor="white",
            edgecolor=color,
            linewidth=linewidth,
        )
    )
    plt.scatter([rect.y], [rect.x], s=size, c=color)


def plot_gradient_field(
    receptive_field_grad: np.ndarray,
    image: np.ndarray = None,
    axis: Optional[Any] = None,
    **plot_params
) -> None:
    """
    Plot gradient map from gradient tensor.

    :param receptive_field_grad: numpy tensor of shape [N, W, H, C]
    :param image: optional image of shape [W, H, 3]
    :param axis: a matplotlib axis object as returned by the e.g. plt.subplot
        function. If not None then axis is used for visualizations otherwise
        default figure is created.
    :param plot_params: additional plot params: figsize=(5, 5)
    """
    receptive_field = estimate_rf_from_gradient(receptive_field_grad)

    receptive_field_grad = np.array(receptive_field_grad).mean(0).mean(-1)
    receptive_field_grad /= receptive_field_grad.max()
    receptive_field_grad += (np.abs(receptive_field_grad) > 0) * 0.2

    if image is not None:
        receptive_field_grad = np.expand_dims(receptive_field_grad, -1)
        receptive_field_grad = 255 / 2 * (receptive_field_grad + 1) + image * 0.5
        receptive_field_grad = receptive_field_grad.astype("uint8")

    if axis is None:
        figsize = plot_params.get("figsize", (5, 5))
        plt.figure(figsize=figsize)
        axis = plt.subplot(111)

    plt.title("Normalized gradient map")
    im = plt.imshow(receptive_field_grad, cmap="coolwarm")
    plt.xlabel("x")
    plt.ylabel("y")

    divider = make_axes_locatable(axis)
    cax = divider.append_axes("right", size="5%", pad=0.05)
    plt.colorbar(im, cax=cax)

    axis.add_patch(
        patches.Rectangle(
            (
                receptive_field.y - receptive_field.h / 2,
                receptive_field.x - receptive_field.w / 2,
            ),  # (x,y)
            receptive_field.h,
            receptive_field.w,
            fill=False,
            alpha=0.9,
            linewidth=4,
            edgecolor=(0.2, 0.2, 0.2),
        )
    )
    axis.set_aspect("equal")
    plt.tight_layout()


def plot_receptive_grid(
    input_shape: GridShape,
    output_shape: GridShape,
    rf_params: ReceptiveFieldDescription,
    custom_image: Optional[np.ndarray] = None,
    plot_naive_rf: bool = False,
    axis: Optional[Any] = None,
    **plot_params
) -> None:
    """
    Visualize receptive field grid.

    :param input_shape: an input image shape as an instance of GridShape
    :param output_shape: an output feature map shape
    :param rf_params: an instance of ReceptiveFieldDescription computed for
        this feature map.
    :param custom_image: optional image [height, width, 3] to be plotted as
        a background.
    :param plot_naive_rf: plot naive version of the receptive field. Naive
        version of RF does not take strides, and offsets into considerations,
        it is a simple linear mapping from N points in feature map to pixels
        in the image.
    :param axis: a matplotlib axis object as returned by the e.g. plt.subplot
        function. If not None then axis is used for visualizations otherwise
        default figure is created.
    :param plot_params: additional plot params: figsize=(5, 5)
    """
    if custom_image is None:
        img = get_default_image(shape=ImageShape(input_shape.h, input_shape.w))
    else:
        img = custom_image

    figsize = plot_params.get("figsize", (10, 10))

    # plot image
    if axis is None:
        plt.figure(figsize=figsize)
        axis = plt.subplot(111)

    axis.imshow(img)
    # plot naive receptive field grid
    if plot_naive_rf:
        dw = input_shape.w / output_shape.w
        dh = input_shape.h / output_shape.h
        for i, j in itertools.product(range(output_shape.w), range(output_shape.h)):
            x0, x1 = i * dw, (i + 1) * dw
            y0, y1 = j * dh, (j + 1) * dh

            axis.add_patch(
                patches.Rectangle(
                    (y0, x0),
                    dh,
                    dw,
                    alpha=0.9,
                    fill=False,
                    edgecolor="gray",
                    linewidth=1,
                )
            )

    rf_offset = rf_params.offset
    rf_size = rf_params.size
    rf_stride = rf_params.stride

    # map from output grid space to input image
    def map_point(i: int, j: int):
        return np.array(rf_offset) + np.array([i, j]) * np.array(rf_stride)

    # plot RF grid based on rf params
    points = [
        map_point(i, j)
        for i, j in itertools.product(range(output_shape.w), range(output_shape.h))
    ]

    points = np.array(points)
    axis.scatter(points[:, 1], points[:, 0], marker="o", c=(0.2, 0.9, 0.1, 0.9), s=10)

    # plot receptive field from corner point
    _plot_rect(
        axis,
        rect=to_rf_rect(rf_offset, rf_size),
        color=(0.9, 0.3, 0.2),
        linewidth=5,
        size=90,
    )
    center_point = map_point(output_shape.w // 2, output_shape.h // 2)
    _plot_rect(
        axis,
        rect=to_rf_rect(GridPoint(center_point[0], center_point[1]), rf_size),
        color=(0.1, 0.3, 0.9),
        linewidth=5,
        size=90,
    )
    last_point = map_point(output_shape.w - 1, output_shape.h - 1)
    _plot_rect(
        axis,
        rect=to_rf_rect(GridPoint(last_point[0], last_point[1]), rf_size),
        color=(0.1, 0.9, 0.3),
        linewidth=5,
        size=90,
    )
    axis.set_aspect("equal")



from typing import Tuple, Callable, List

import numpy as np
import tensorflow as tf

from receptivefield.base import ReceptiveField
from receptivefield.logging import get_logger
from receptivefield.types import ImageShape, GridShape, GridPoint, \
    FeatureMapDescription

_logger = get_logger()


def _get_tensor_shape(tensor: tf.Tensor) -> List[int]:
    """
    Parse TensorShape to python tuple:
    Example: TensorShape([
        Dimension(1), Dimension(8), Dimension(8), Dimension(512)
    ]) will return [1, 8, 8, 32]
    :param tensor: tensorflow Tensor
    :return: integer type tuple
    """
    return list(map(int, tensor.get_shape()))


def _define_fm_gradient(
        input_tensor: tf.Tensor,
        output_tensor: tf.Tensor,
        receptive_field_mask: tf.Tensor
) -> tf.Tensor:
    """
    Define gradient of feature map w.r.t. the input image.

    :param input_tensor: an input image tensor [h, w, 3]
    :param output_tensor: an feature map tensor [fm_h, fm_w, num_channels]
    :param receptive_field_mask: a backpropagation mask [fm_h, fm_w, 1]

    :return: a gradient tensor [h, w, 3]
    """
    x = tf.reduce_mean(output_tensor, -1, keep_dims=True)
    fake_loss = x * receptive_field_mask
    fake_loss = tf.reduce_mean(fake_loss)
    # define gradient w.r.t. image
    return tf.gradients(fake_loss, input_tensor)[0]


def _get_gradient_from_grid_points(
        rf_extractor: ReceptiveField,
        points: List[GridPoint],
        intensity: float = 1.0
) -> List[np.ndarray]:
    """
    Computes gradient at input image tenor generated by
    point-like perturbation at output grid location given by
    @point coordinates.

    :param rf_extractor: and instance of TFReceptiveField or
        TFFeatureMapsReceptiveField.
    :param points: source coordinate of the backpropagated gradient for each
        feature map.
    :param intensity: scale of the gradient, default = 1
    :return gradient maps for each feature map
    """

    input_shape = rf_extractor._input_shape.replace(n=1)
    output_feature_maps = []
    for fm in range(rf_extractor.num_feature_maps):
        output_shape = rf_extractor._output_shapes[fm].replace(n=1)
        output_feature_map = np.zeros(shape=output_shape)
        output_feature_map[:, points[fm].y, points[fm].x, 0] = intensity
        output_feature_maps.append(output_feature_map)

    return rf_extractor._gradient_function(
        output_feature_maps, np.zeros(shape=input_shape)
    )


class TFReceptiveField(ReceptiveField):
    def __init__(self, model_func: Callable[[ImageShape], tf.Tensor]):
        """
        :param model_func: model creation function. Function which accepts image
            shape [H, W, C] and returns tensorflow graph.
        """
        self._session: tf.Session = None
        super().__init__(model_func)

    def _prepare_gradient_func(
        self, input_shape: ImageShape, input_tensor: str, output_tensors: List[str]
    ) -> Tuple[Callable, GridShape, List[GridShape]]:
        """
        Computes gradient function and additional parameters. Note
        that the receptive field parameters like stride or size, do not
        depend on input image shape. However, if the RF of original network
        is bigger than input_shape this method will fail. Hence it is
        recommended to increase the input shape.

        :param input_shape: shape of the input image. Used in @model_func.
        :param input_tensor: name of the input image tensor.
        :param output_tensors: a list of names of the target
            feature map tensors.

        :returns
            gradient_function: a function which returns gradient w.r.t. to
                the input image.
            input_shape: a shape of the input image tensor.
            output_shape: a list shapes of the output feature map tensors.
        """

        if self._session is not None:
            tf.reset_default_graph()
            self._session.close()

        with tf.Graph().as_default() as graph:
            with tf.variable_scope("", reuse=tf.AUTO_REUSE):

                # this function will create default graph
                _ = self._model_func(ImageShape(*input_shape))

                # default_graph = tf.get_default_graph()
                # get graph tensors by names
                input_tensor = graph.get_operation_by_name(input_tensor).outputs[0]
                input_shape = _get_tensor_shape(input_tensor)

                grads = []
                receptive_field_masks = []
                output_shapes = []

                for output_tensor in output_tensors:
                    output_tensor = graph.get_operation_by_name(output_tensor).outputs[0]

                    # shapes
                    output_shape = _get_tensor_shape(output_tensor)
                    output_shape = (1, output_shape[1], output_shape[2], 1)
                    output_shapes.append(output_shape)

                    # define loss function
                    receptive_field_mask = tf.placeholder(
                        tf.float32, shape=output_shape, name="grid"
                    )
                    grad = _define_fm_gradient(
                        input_tensor, output_tensor, receptive_field_mask
                    )
                    grads.append(grad)
                    receptive_field_masks.append(receptive_field_mask)

                _logger.info(f"Feature maps shape: {output_shapes}")
                _logger.info(f"Input shape       : {input_shape}")

            self._session = tf.Session(graph=graph)
            self._session.run(tf.global_variables_initializer())

            def gradient_fn(fm_masks, input_image):
                fetch_dict = {
                    mask_t: mask_np
                    for mask_t, mask_np in zip(receptive_field_masks, fm_masks)
                }
                fetch_dict[input_tensor] = input_image
                return self._session.run(grads, feed_dict=fetch_dict)

        return (
            gradient_fn,
            GridShape(*input_shape),
            [GridShape(*output_shape) for output_shape in output_shapes],
        )

    def _get_gradient_from_grid_points(
        self, points: List[GridPoint], intensity: float = 1.0
    ) -> List[np.ndarray]:
        """
        Computes gradient at image tensor generated by
        point-like perturbation at output grid location given by
        @point coordinates.

        :param points: source coordinate of the backpropagated gradient for each
            feature map.
        :param intensity: scale of the gradient, default = 1
        :return gradient maps for each feature map
        """

        return _get_gradient_from_grid_points(
            self,
            points=points,
            intensity=intensity
        )

    def compute(
            self,
            input_shape: ImageShape,
            input_tensor: str,
            output_tensors: List[str]
    ) -> List[FeatureMapDescription]:

        """
        Compute ReceptiveFieldDescription of given model for image of
        shape input_shape [H, W, C]. If receptive field of the network
        is bigger thant input_shape this method will raise exception.
        In order to solve with problem try to increase input_shape.

        :param input_shape: shape of the input image e.g. (224, 224, 3)
        :param input_tensor: name of the input tensor
        :param output_tensors: a list of names of the target feature map tensors.

        :return a list of estimated FeatureMapDescription for each feature
            map.
        """

        return super().compute(
            input_shape=input_shape,
            input_tensor=input_tensor,
            output_tensors=output_tensors
        )


class TFFeatureMapsReceptiveField(ReceptiveField):
    def __init__(self, model_func: Callable[[tf.Tensor], List[tf.Tensor]]):
        """
        :param model_func: model creation function. Function which
            accepts image_tensor of shape [1, H, W, C] as an input and
            returns list of tensors which correspond to selected feature maps.
        """
        self._session = None
        super().__init__(model_func)

    def _prepare_gradient_func(
        self, input_shape: ImageShape
    ) -> Tuple[Callable, GridShape, List[GridShape]]:

        """
        Computes gradient function and additional parameters. Note
        that the receptive field parameters like stride or size, do not
        depend on input image shape. However, if the RF of original network
        is bigger than input_shape this method will fail. Hence it is
        recommended to increase the input shape.

        :param input_shape: shape of the input image. Used in @model_func.

        :returns
            gradient_function: a function which returns gradient w.r.t. to
                the input image
            input_shape: a shape of the input image tensor
            output_shapes: a list shapes of the output feature map tensors
        """

        if self._session is not None:
            tf.reset_default_graph()
            self._session.close()

        with tf.Graph().as_default() as graph:
            with tf.variable_scope("", reuse=tf.AUTO_REUSE):

                input_tensor = tf.placeholder(
                    tf.float32, shape=[1, *input_shape], name="input_image"
                )
                input_shape = _get_tensor_shape(input_tensor)
                feature_maps = self._model_func(input_tensor)

                grads = []
                receptive_field_masks = []
                output_shapes = []

                for output_tensor in feature_maps:
                    # shapes
                    output_shape = _get_tensor_shape(output_tensor)
                    output_shape = (1, output_shape[1], output_shape[2], 1)
                    output_shapes.append(output_shape)

                    # define loss function
                    receptive_field_mask = tf.placeholder(
                        tf.float32, shape=output_shape, name="grid"
                    )
                    grad = _define_fm_gradient(
                        input_tensor, output_tensor, receptive_field_mask
                    )
                    grads.append(grad)
                    receptive_field_masks.append(receptive_field_mask)

            _logger.info(f"Feature maps shape: {output_shapes}")
            _logger.info(f"Input shape       : {input_shape}")

            self._session = tf.Session(graph=graph)
            self._session.run(tf.global_variables_initializer())

            def gradient_fn(
                fm_masks: List[np.ndarray], input_image: np.ndarray
            ) -> List[np.ndarray]:
                fetch_dict = {
                    mask_t: mask_np
                    for mask_t, mask_np in zip(receptive_field_masks, fm_masks)
                }
                fetch_dict[input_tensor] = input_image
                return self._session.run(grads, feed_dict=fetch_dict)

        return (
            gradient_fn,
            GridShape(*input_shape),
            [GridShape(*output_shape) for output_shape in output_shapes],
        )

    def _get_gradient_from_grid_points(
            self, points: List[GridPoint], intensity: float = 1.0
    ) -> List[np.ndarray]:
        """
        Computes gradient at input image tensor generated by
        point-like perturbation at output grid location given by
        @point coordinates.

        :param points: source coordinate of the backpropagated gradient for each
            feature map.
        :param intensity: scale of the gradient, default = 1
        :return gradient maps for each feature map
        """
        return _get_gradient_from_grid_points(
            self,
            points=points,
            intensity=intensity
        )

    def compute(self, input_shape: ImageShape) -> List[FeatureMapDescription]:
        """
        Compute ReceptiveFieldDescription of given model for image of
        shape input_shape [H, W, C]. If receptive field of the network
        is bigger thant input_shape this method will raise exception.
        In order to solve with problem try to increase input_shape.

        :param input_shape: shape of the input image e.g. (224, 224, 3)

        :return a list of estimated FeatureMapDescription for each feature
            map.
        """
        return super().compute(input_shape=input_shape)



from typing import Tuple, Callable, List

import numpy as np
import torch
import torch.nn as nn

from receptivefield.base import ReceptiveField
from receptivefield.logging import get_logger
from receptivefield.types import (
    ImageShape,
    GridPoint,
    GridShape,
    FeatureMapDescription,
)

_logger = get_logger()


def _define_receptive_field_func(
    model_fn: Callable[[], nn.Module], input_shape: GridShape
):
    shape = [input_shape.n, input_shape.c, input_shape.h, input_shape.w]
    input_image = torch.zeros(*shape)
    model = model_fn()
    _ = model(input_image)

    feature_maps = model.__dict__.get('feature_maps', None)
    if feature_maps is None:
        raise ValueError('Field module.feature_maps is not defined. '
                         'Cannot compute receptive fields.')

    if type(feature_maps) != list:
        raise ValueError('Field module.feature_maps must be a list. '
                         'Cannot compute receptive fields.')

    # compute feature maps output shapes
    output_shapes = []
    for feature_map in feature_maps:
        output_shape = feature_map.size()
        output_shape = GridShape(
            n=output_shape[0],
            c=output_shape[1],
            h=output_shape[2],
            w=output_shape[3]
        )
        output_shapes.append(output_shape)

    _logger.info(f"Feature maps shape: {output_shapes}")
    _logger.info(f"Input shape       : {input_shape}")

    def gradient_function(
            receptive_field_masks: List[torch.Tensor]
    ) -> List[np.ndarray]:

        grads = []
        for fm, rf_mask in enumerate(receptive_field_masks):
            input_tensor = torch.zeros(*shape)
            input_tensor = torch.autograd.Variable(
                input_tensor, requires_grad=True
            )
            model.zero_grad()
            _ = model(input_tensor)

            fm = torch.mean(model.feature_maps[fm], 1, keepdim=True)
            fake_loss = fm * rf_mask
            fake_loss = torch.mean(fake_loss)
            fake_loss.backward()
            grads.append(input_tensor.grad.detach().numpy())

        return grads

    return gradient_function, input_shape, output_shapes


class PytorchReceptiveField(ReceptiveField):
    def __init__(self, model_func: Callable[[], nn.Module]):
        """
        :param model_func: model function which returns new instance of
            nn.Module.
        """
        super().__init__(model_func)

    def _prepare_gradient_func(
        self, input_shape: ImageShape
    ) -> Tuple[Callable, GridShape, List[GridShape]]:
        """
        Computes gradient function and additional parameters. Note
        that the receptive field parameters like stride or size, do not
        depend on input image shape. However, if the RF of original network
        is bigger than input_shape this method will fail. Hence it is
        recommended to increase the input shape.

        :param input_shape: shape of the input image, which is feed to the
            Pytorch module.

        :returns
            gradient_function: a function which returns gradient w.r.t. to
                the input image.
            input_shape: a shape of the input image tensor.
            output_shapes: a list shapes of the output feature map tensors.
        """
        input_shape = ImageShape(*input_shape)
        input_shape = GridShape(
            n=1,
            c=input_shape.c,
            h=input_shape.h,
            w=input_shape.w,
        )

        gradient_function, input_shape, output_shapes = \
            _define_receptive_field_func(self._model_func, input_shape)

        return gradient_function, input_shape, output_shapes

    def _get_gradient_from_grid_points(
        self, points: List[GridPoint], intensity: float = 1.0
    ) -> List[np.ndarray]:
        """
        Computes gradient at image tensor generated by
        point-like perturbation at output grid location given by
        @point coordinates.

        :param points: source coordinate of the backpropagated gradient for each
            feature map.
        :param intensity: scale of the gradient, default = 1
        :return a list gradient maps of shape [1, C, H, W] for each feature map
        """
        output_feature_maps = []
        for fm in range(self.num_feature_maps):
            os = self._output_shapes[fm]
            # input tensors are in NCHW format
            output_feature_map = torch.zeros(size=[1, 1, os.h, os.w])
            output_feature_map[:, :, points[fm].y, points[fm].x] = intensity
            output_feature_maps.append(output_feature_map)

        def _postprocess_grad(grad: np.ndarray) -> np.ndarray:
            # convert to NHWC format
            grad = np.abs(np.transpose(grad, [0, 2, 3, 1]))
            return grad / grad.max()

        torch_grads = self._gradient_function(output_feature_maps)
        return [_postprocess_grad(grad) for grad in torch_grads]

    def compute(self, input_shape: ImageShape) -> List[FeatureMapDescription]:
        """
        Compute ReceptiveFieldDescription of given model for image of
        shape input_shape [H, W, C]. If receptive field of the network
        is bigger thant input_shape this method will raise exception.
        In order to solve this problem try to increase input_shape.

        :param input_shape: shape of the input image e.g. (224, 224, 3)

        :return a list of estimated FeatureMapDescription for each feature
            map.
        """
        return super().compute(input_shape=input_shape)



from abc import ABCMeta, abstractmethod
from typing import Tuple, Callable, Any, Optional, List
import matplotlib.pyplot as plt
import numpy as np

from receptivefield.common import estimate_rf_from_gradients
from receptivefield.logging import get_logger
from receptivefield.plotting import plot_receptive_grid, plot_gradient_field
from receptivefield.types import (
    ImageShape,
    GridPoint,
    GridShape,
    ReceptiveFieldDescription,
    Size,
    FeatureMapDescription,
)

_logger = get_logger()


class ReceptiveField(metaclass=ABCMeta):
    def __init__(self, model_func: Callable[[Any], Any]) -> None:

        self._model_func: Callable[[Any], Any] = model_func
        self._gradient_function: Callable = None
        self._input_shape: GridShape = None
        self._output_shapes: List[GridShape] = None
        self._rf_params: List[ReceptiveFieldDescription] = None
        self._built: bool = False

    @property
    def feature_maps_desc(self) -> List[FeatureMapDescription]:
        """Return description of all feature maps"""
        self._check()
        return [
            FeatureMapDescription(size=Size(size.w, size.h), rf=rf)
            for size, rf in zip(self._output_shapes, self._rf_params)
        ]

    @property
    def input_shape(self) -> ImageShape:
        """Return input shape of the feature extractor"""
        self._check()
        return ImageShape(
            w=self._input_shape.w, h=self._input_shape.h, c=self._input_shape.c
        )

    @property
    def output_shapes(self) -> List[Size]:
        """Return a list of sizes of selected features maps"""
        self._check()
        return [Size(w=size.w, h=size.h) for size in self._output_shapes]

    @property
    def num_feature_maps(self) -> int:
        """Returns number of features maps"""
        self._check()
        return len(self._output_shapes)

    def _build_gradient_func(self, *args, **kwargs) -> None:
        """
        Build gradient function and collect the input shape of the image
        and feature map shapes.

        :param args: a list of arguments which depend on the API
        :param kwargs: keywords which depend on the API
        """
        gradient_function, input_shape, output_shapes = \
            self._prepare_gradient_func(*args, **kwargs)

        self._built = True
        self._gradient_function = gradient_function
        self._input_shape = input_shape
        self._output_shapes = output_shapes

    @abstractmethod
    def _prepare_gradient_func(
        self, *args, **kwargs
    ) -> Tuple[Callable, GridShape, List[GridShape]]:
        """
        Computes gradient function and additional parameters. Note
        that the receptive field parameters like stride or size, do not
        depend on input image shape. However, if the RF of original network
        is bigger than input_shape this method will fail. Hence it is
        recommended to increase the input shape.

        The arguments args, and kwargs must be compatible with args in `compute`
        and `_build_gradient_func` functions.

        :param args: a list of arguments which depend on the API
        :param kwargs: keywords which depend on the API

        :returns
            gradient_function: a function which returns gradient w.r.t. to
                the input image. The usage of the gradient function may depend
                on the API.
            input_shape: a shape of the input image tensor
            output_shape: a shapes of the output feature map tensors
        """
        pass

    @abstractmethod
    def _get_gradient_from_grid_points(
        self, points: List[GridPoint], intensity: float = 1.0
    ) -> List[np.ndarray]:
        """
        Computes gradient at input image generated by
        point-like perturbation at output grid location given by
        @point coordinates.

        :param points: source coordinate of the backpropagated gradient for each
            feature map.
        :param intensity: scale of the gradient, default = 1
        :return gradient maps for each feature map
        """
        pass

    def _get_gradient_activation_at_map_center(
        self, center_offsets: List[GridPoint], intensity: float = 1
    ) -> List[np.ndarray]:
        points = []
        for fm in range(self.num_feature_maps):
            output_shape = self._output_shapes[fm]
            center_offset = center_offsets[fm]

            _logger.debug(
                f"Computing receptive field for feature map [{fm}] at center "
                f"({output_shape.w//2}, {output_shape.h//2}) "
                f"with offset {center_offset}"
            )

            # compute grid center
            w = output_shape.w
            h = output_shape.h
            cx = w // 2 - 1 if w % 2 == 0 else w // 2
            cy = h // 2 - 1 if h % 2 == 0 else h // 2

            cx += center_offset.x
            cy += center_offset.y
            points.append(GridPoint(x=cx, y=cy))
        return self._get_gradient_from_grid_points(points=points, intensity=intensity)

    def _check(self):
        if not self._built:
            raise Exception(
                "Receptive field not computed. Run compute function.")

    def compute(self, *args, **kwargs) -> List[FeatureMapDescription]:
        """
        Compute ReceptiveFieldDescription of given model for image of
        shape input_shape [H, W, C]. If receptive field of the network
        is bigger thant input_shape this method will raise exception.
        In order to solve with problem try to increase input_shape.

        :param args: a list of arguments which depend on the API
        :param kwargs: keywords which depend on the API

        :return a list of estimated FeatureMapDescription for each feature
            map.
        """
        # define gradient function
        self._build_gradient_func(*args, **kwargs)

        # receptive field at map center
        rf_grads00 = self._get_gradient_activation_at_map_center(
            center_offsets=[GridPoint(0, 0)] * self.num_feature_maps
        )
        rfs_at00 = estimate_rf_from_gradients(rf_grads00)

        # receptive field at map center with offset (1, 1)
        rf_grads11 = self._get_gradient_activation_at_map_center(
            center_offsets=[GridPoint(1, 1)] * self.num_feature_maps
        )
        rfs_at11 = estimate_rf_from_gradients(rf_grads11)

        # receptive field at feature map grid start x=0, y=0
        rf_grads_point00 = self._get_gradient_from_grid_points(
            points=[GridPoint(0, 0)] * self.num_feature_maps
        )
        rfs_at_point00 = estimate_rf_from_gradients(rf_grads_point00)

        self._rf_params = []

        for fm, (rf_at_point00, rf_at00, rf_at11) in enumerate(
            zip(rfs_at_point00, rfs_at00, rfs_at11)
        ):
            # compute position of the first anchor, center point of rect
            x0 = rf_at_point00.w - rf_at00.w / 2
            y0 = rf_at_point00.h - rf_at00.h / 2

            # compute feature map/input image offsets
            dx = rf_at11.x - rf_at00.x
            dy = rf_at11.y - rf_at00.y

            # compute receptive field size
            size = Size(rf_at00.w, rf_at00.h)

            rf_params = ReceptiveFieldDescription(
                offset=(x0, y0), stride=(dx, dy), size=size
            )

            _logger.info(
                f"Estimated receptive field for feature map [{fm}]: {rf_params}"
            )
            self._rf_params.append(rf_params)

        return self.feature_maps_desc

    def plot_gradients_at(
        self,
        points: List[GridPoint],
        image: Optional[np.ndarray] = None,
        layout: Optional[Tuple[int, int]] = None,
        figsize: Optional[Tuple[int, int]] = None,
    ) -> None:
        """
        Plot gradient map generated by certain `point` at feature map.

        :param points: a list of  GridPoint on feature maps.
        :param image: (optional) image of shape [W, H, 3]
        :param layout: (optional) a matplotlib subplot layout given by
            tuple of form (num_rows, num_cols). If None the layout is set
            to (num_feature_maps, 1).
        :param figsize: (optional) a matplotlib figsize. If None the default
            figure is used.
        """

        if layout is None:
            layout = (self.num_feature_maps, 1)
        if figsize is not None:
            plt.figure(figsize=figsize)

        receptive_field_grads = self._get_gradient_from_grid_points(
            points=[GridPoint(*point) for point in points]
        )

        for fm in range(self.num_feature_maps):
            axis = plt.subplot(layout[0], layout[1], fm + 1)
            plot_gradient_field(
                receptive_field_grad=receptive_field_grads[fm], image=image, axis=axis
            )

    def plot_gradient_at(
        self,
        fm_id: int,
        point: GridPoint,
        image: Optional[np.ndarray] = None,
        **plot_params,
    ) -> None:
        """
        Plot gradient map generated by certain `point` at selected feature map.

        :param fm_id: an index of feature map for which gradient map
            will be plotted.
        :param point: a GridPoint on feature map selected by fm_id.
        :param image: (optional) image of shape [W, H, 3]
        :param plot_params: (optional) additional plot parameters:
            figsize: tuple of int (5, 5)
            axis: a matplotlib axis object as returned by e.g. plt.subplot
                function. If not None then provided axis is used for
                visualizations otherwise default figure is created, with
                optional figsize.
        """
        points = [GridPoint(0, 0)] * self.num_feature_maps
        points[fm_id] = GridPoint(*point)
        receptive_field_grads = self._get_gradient_from_grid_points(points=points)

        plot_gradient_field(
            receptive_field_grad=receptive_field_grads[fm_id],
            image=image,
            **plot_params,
        )

    def plot_rf_grid(
        self,
        fm_id: int,
        custom_image: Optional[np.ndarray] = None,
        plot_naive_rf: bool = False,
        **plot_params,
    ) -> None:
        """
        Visualize receptive field grid for selected feature map.

        :param fm_id: an index of feature map for which gradient map
            will be plotted.
        :param custom_image: (optional) image of shape [W, H, 3]
        :param plot_naive_rf: plot naive version of the receptive field. Naive
            version of RF does not take strides, and offsets into considerations,
            it is a simple linear mapping from N points in feature map to pixels
            in the image.
        :param plot_params: (optional) additional plot parameters:
            figsize: tuple of int (5, 5)
            axis: a matplotlib axis object as returned by e.g. plt.subplot
                function. If not None then provided axis is used for
                visualizations otherwise default figure is created, with
                optional figsize.
        """
        plot_receptive_grid(
            input_shape=self._input_shape,
            output_shape=self._output_shapes[fm_id],
            rf_params=self._rf_params[fm_id],
            custom_image=custom_image,
            plot_naive_rf=plot_naive_rf,
            **plot_params,
        )

    def plot_rf_grids(
        self,
        custom_image: Optional[np.ndarray] = None,
        plot_naive_rf: bool = False,
        layout: Optional[Tuple[int, int]] = None,
        figsize: Optional[Tuple[int, int]] = None,
    ) -> None:
        """
        Visualize receptive field grid for selected feature map.

        :param custom_image: (optional) image of shape [W, H, 3]
        :param plot_naive_rf: plot naive version of the receptive field. Naive
            version of RF does not take strides, and offsets into considerations,
            it is a simple linear mapping from N points in feature map to pixels
            in the image.
        :param layout: (optional) a matplotlib subplot layout given by
            tuple of form (num_rows, num_cols). If None the layout is set
            to (num_feature_maps, 1).
        :param figsize: (optional) a matplotlib figsize. If None the default
            figure is used.
        """

        if layout is None:
            layout = (self.num_feature_maps, 1)
        if figsize is not None:
            plt.figure(figsize=figsize)

        for fm in range(self.num_feature_maps):
            axis = plt.subplot(layout[0], layout[1], fm + 1)
            self.plot_rf_grid(
                fm_id=fm,
                custom_image=custom_image,
                plot_naive_rf=plot_naive_rf,
                axis=axis,
            )



import os
from typing import Union, List

import numpy
from PIL import Image

from receptivefield.types import ImageShape

dir_path = os.path.dirname(os.path.realpath(__file__))


def get_default_images() -> List[str]:
    """
    List available image names, included in receptive field library.
    Those names can be used in get_default_image() function.

    :return: list of available names
    """
    images = os.listdir(os.path.join(dir_path, "resources"))
    return [im.replace(".jpg", "") for im in images if "jpg" in im]


def _get_default_image_path(name: str) -> str:

    if name not in get_default_images():
        raise Exception(
            f"Image name is '{name}' but should be one of "
            f"the following: {get_default_images()}"
        )

    image_path = os.path.join(dir_path, f"resources/{name}.jpg")
    if not os.path.exists(image_path):
        raise FileNotFoundError(f"Image {name} does not exist :(")

    return image_path


def tile_numpy_image(
    image: numpy.ndarray, tile_factor: int = 0, shade: bool = True
) -> numpy.ndarray:
    """
    Tile array of shape [W, H, C], where C=3.

    :param image: image array
    :param tile_factor: number of tiles, if tile_factor=0, this
        function returns a copy of the input image
    :param shade: make non-central tiles gray scaled.
    :return: numpy array image of shape [W*n, H*n, C],
        where n=2 * tile_factor + 1
    """
    timage = tile_pil_image(
        Image.fromarray(numpy.uint8(image)), tile_factor, shade
    )
    return numpy.array(timage)


def tile_pil_image(
    image: Image.Image, tile_factor: int = 0, shade: bool = True
) -> Image.Image:
    """
    Tile PIL Image.

    :param image: PIL Image object
    :param tile_factor: number of tiles, if tile_factor=0, this
        function returns a copy of the input image
    :param shade: make non-central tiles gray scaled.
    :return: Image of shape [W*n, H*n, C], where n=2 * tile_factor + 1
    """
    shape = ImageShape(*image.size)
    new_img = image.copy()
    if tile_factor > 0:
        tf = 2 * tile_factor + 1
        new_img = Image.new("RGB", (shape.w * tf, shape.h * tf))
        new_shape = ImageShape(*new_img.size)
        gray_img = image.convert("LA").convert("RGB")

        for n, i in enumerate(range(0, new_shape.w, shape.w)):
            for m, j in enumerate(range(0, new_shape.h, shape.h)):
                distance = (abs(m - tile_factor) + abs(n - tile_factor)) / tf
                alpha = distance > 0 if shade else 0
                # place image at position (i, j)
                new_img.paste(Image.blend(image, gray_img, alpha), (i, j))
    return new_img


def get_default_image(
    shape: ImageShape,
    tile_factor: int = 0,
    shade: bool = True,
    as_image: bool = False,
    name: str = "lena",
) -> Union[numpy.ndarray, Image.Image]:
    """
    Loads default image from resources and reshape it to size
    shape.

    :param shape: [width, height]
    :param tile_factor: tile image, if 0 the resulting image shape is
        [width, height], otherwise the output size is defined by number of
        tiles. tile_factor is a non-negative integer number.
    :param shade: if True and tile_factor > 0 it makes tiles gray scale
    :param as_image: if True, function returns PIL Image object, else
        numpy array.
    :param name: name of the default image to be loaded. Call
        get_default_images() to list available images names (default - lena).

    :return: numpy array of shape [width, height, 3] if as_image=False,
        otherwise PIL Image object
    """
    shape = ImageShape(*shape)
    tile_factor = int(tile_factor)

    img = Image.open(_get_default_image_path(name=name), mode="r")
    img = img.resize((shape.w, shape.h), Image.ANTIALIAS)
    img = tile_pil_image(img, tile_factor=tile_factor, shade=shade)

    if as_image:
        return img
    return numpy.array(img)



from typing import Tuple, List

import numpy as np

from receptivefield.logging import get_logger
from receptivefield.types import GridShape, ReceptiveFieldRect

_logger = get_logger()


def _compute_fans(shape: GridShape) -> Tuple[int, int]:
    """Computes the number of input and output units for a weight shape.

    # Arguments
        shape: Integer shape tuple of type GridShape (N, W, H, C)

    # Returns
        A tuple of scalars, `(fan_in, fan_out)`.
    """
    if len(shape) == 2:
        fan_in = shape[0]
        fan_out = shape[1]
    elif len(shape) in {3, 4, 5}:
        receptive_field_size = np.prod(shape[:-2])
        fan_in = shape[-2] * receptive_field_size
        fan_out = shape[-1] * receptive_field_size
    else:
        # No specific assumptions.
        fan_in = np.sqrt(np.prod(shape))
        fan_out = np.sqrt(np.prod(shape))
    return fan_in, fan_out


def scaled_constant(
    scale: float, shape: GridShape, mode: str = "fan_avg", dtype: type = np.float32
) -> np.ndarray:
    """
    Returns np array of given @shape filed with constant value.
    :param scale: constant value multiplier, default 1
    :param shape: shape of resulting tensor
    :param mode: type of constant value estimation.
    :param dtype:
    :return: numpy tensor
    """
    fan_in, fan_out = _compute_fans(shape)
    if mode == "fan_in":
        scale /= max(1., fan_in)
    elif mode == "fan_out":
        scale /= max(1., fan_out)
    elif mode == "fan_avg":
        scale /= max(1., float(fan_in + fan_out) / 2)

    limit = scale
    return limit * np.ones(shape, dtype=dtype)


def estimate_rf_from_gradient(receptive_field_grad: np.ndarray) -> ReceptiveFieldRect:
    """
    Given input gradient tensors of shape [N, W, H, C] it returns the
    estimated size of gradient `blob` in W-H directions i.e. this
    function computes the size of gradient in W-H axis for each feature map.

    :param receptive_field_grad: a numpy tensor with gradient values
        obtained for certain feature map
    :return: a corresponding ReceptiveFieldRect
    """

    receptive_field_grad = np.array(receptive_field_grad).mean(0).mean(-1)
    binary_map: np.ndarray = (receptive_field_grad[:, :] > 0)

    x_cs: np.ndarray = binary_map.sum(-1) >= 1
    y_cs: np.ndarray = binary_map.sum(0) >= 1

    x = np.arange(len(x_cs))
    y = np.arange(len(y_cs))

    width = x_cs.sum()
    height = y_cs.sum()

    x = np.sum(x * x_cs) / width
    y = np.sum(y * y_cs) / height

    return ReceptiveFieldRect(x, y, width, height)


def estimate_rf_from_gradients(
    receptive_field_grads: List[np.ndarray]
) -> List[ReceptiveFieldRect]:
    """
    Given input gradient tensors of shape [N, W, H, C] it returns the
    estimated size of gradient `blob` in W-H directions i.e. this
    function computes the size of gradient in W-H axis for each feature map.

    :param receptive_field_grads: a list of numpy tensor with gradient values
        obtained for different feature maps
    :return: a list of corresponding ReceptiveFieldRect
    """

    return [
        estimate_rf_from_gradient(receptive_field_grad)
        for receptive_field_grad in receptive_field_grads
    ]



from typing import Tuple, Callable, Any, List

import keras.backend as K
import numpy as np
from keras.engine import Layer
from keras.layers import Conv2D, MaxPool2D, Activation
from keras.layers import Input, Lambda
from keras.models import Model
import keras.layers as layers

from receptivefield.base import ReceptiveField
from receptivefield.common import scaled_constant
from receptivefield.logging import get_logger
from receptivefield.types import ImageShape, GridPoint, GridShape, \
    FeatureMapDescription

_logger = get_logger()


def _check_activation(layer: Layer):
    if issubclass(type(layer), Layer):
        layer_act = layer.activation.__name__
    else:
        layer_act = None

    if layer_act != 'linear':
        _logger.warning(f"Layer {layer.name} activation should be linear "
                        f"but is: {layer_act}")


def setup_model_weights(model: Model) -> None:
    """
    Set all weights to be a constant values. Biases are set to zero.
    Only Conv2D are supported.

    :param model: a Keras model
    """
    _logger.info(f"Running `setup_model_weights` on model: {model}")
    for layer in model.layers:
        # check layer type
        if type(layer) == MaxPool2D:
            _logger.warning(f"MaxPool2D detected: {layer.name}. Replace it with"
                            f" AvgPool2D in order to obtain better receptive "
                            f"field mapping estimation")

        if type(layer) in [layers.AvgPool2D, layers.InputLayer]:
            continue

        # set weights
        if type(layer) == Conv2D:
            _logger.debug(f"Setting weights for layer {layer.name}")
            weights = layer.get_weights()
            w_kernel = scaled_constant(1, weights[0].shape)
            w = [w_kernel]
            if len(weights) > 1:
                w_bias = np.zeros_like(weights[1])
                w.append(w_bias)
            layer.set_weights(w)
            _check_activation(layer)
        elif type(layer) == Activation:
            _check_activation(layer)
        else:
            _logger.warning(f"Setting weights for layer {type(layer)} "
                            f"is not supported.")


def _define_receptive_field_func(
        model: Model,
        input_layer: str,
        feature_map_layers: List[str]
):

    output_shapes = []
    input_shape = model.get_layer(input_layer).output_shape
    grads = []
    receptive_field_masks = []

    for fm in feature_map_layers:
        output_shape = model.get_layer(fm).output_shape
        output_shape = [*output_shape[:-1], 1]
        output_shapes.append(output_shape)

        receptive_field_mask = Input(shape=output_shape[1:])
        x = model.get_layer(fm).output
        x = Lambda(lambda _x: K.mean(_x, -1, keepdims=True))(x)

        fake_loss = x * receptive_field_mask
        fake_loss = K.mean(fake_loss)
        grad = K.gradients(fake_loss, model.input)

        grads.append(grad[0])
        receptive_field_masks.append(receptive_field_mask)

    _logger.info(f"Feature maps shape: {output_shapes}")
    _logger.info(f"Input shape       : {input_shape}")

    gradient_function = K.function(
        inputs=[*receptive_field_masks, model.input, K.learning_phase()],
        outputs=grads)

    return gradient_function, input_shape, output_shapes


class KerasReceptiveField(ReceptiveField):
    def __init__(
            self,
            model_func: Callable[[ImageShape], Any],
            init_weights: bool = False
    ):
        """
        Build Keras receptive field estimator.

        :param model_func: model creation function
        :param init_weights: if True all conv2d weights are overwritten
        by constant value.
        """
        super().__init__(model_func)
        self.init_weights = init_weights

    def _prepare_gradient_func(
            self,
            input_shape: ImageShape,
            input_layer: str,
            output_layers: List[str]
    ) -> Tuple[Callable, GridShape, List[GridShape]]:
        """
        Computes gradient function and additional parameters. Note
        that the receptive field parameters like stride or size, do not
        depend on input image shape. However, if the RF of original network
        is bigger than input_shape this method will fail. Hence it is
        recommended to increase the input shape.

        :param input_shape: shape of the input image. Used in @model_func.
        :param input_layer: name of the input layer.
        :param output_layers: a list of names of the target feature map layers.

        :returns
            gradient_function: a function which returns gradient w.r.t. to
                the input image
            input_shape: a shape of the input image tensor
            output_shape: a shapes of the output feature map tensors
        """
        model = self._model_func(ImageShape(*input_shape))
        if self.init_weights:
            setup_model_weights(model)

        gradient_function, input_shape, output_shapes = \
            _define_receptive_field_func(model, input_layer, output_layers)

        return gradient_function, GridShape(*input_shape), \
            [GridShape(*output_shape) for output_shape in output_shapes]

    def _get_gradient_from_grid_points(
            self,
            points: List[GridPoint],
            intensity: float = 1.0
    ) -> List[np.ndarray]:
        """
        Computes gradient at input_layer (image_layer) generated by
        point-like perturbation at output grid location given by
        @point coordinates.

        :param points: source coordinate of the backpropagated gradient for each
            feature map.
        :param intensity: scale of the gradient, default = 1
        :return gradient maps for each feature map
        """
        input_shape = self._input_shape.replace(n=1)
        output_feature_maps = []
        for fm in range(self.num_feature_maps):
            output_shape = self._output_shapes[fm].replace(n=1)
            output_feature_map = np.zeros(shape=output_shape)
            output_feature_map[:, points[fm].y, points[fm].x, 0] = intensity
            output_feature_maps.append(output_feature_map)

        receptive_field_grads = self._gradient_function([
            *output_feature_maps, np.zeros(shape=input_shape), 0])

        return receptive_field_grads

    def compute(
            self,
            input_shape: ImageShape,
            input_layer: str,
            output_layers: List[str]
    ) -> List[FeatureMapDescription]:

        """
        Compute ReceptiveFieldDescription of given model for image of
        shape input_shape [H, W, C]. If receptive field of the network
        is bigger thant input_shape this method will raise exception.
        In order to solve with problem try to increase input_shape.

        :param input_shape: shape of the input image e.g. (224, 224, 3)
        :param input_layer: name of the input layer
        :param output_layers: a list of names of the target feature map layers.

        :return a list of estimated FeatureMapDescription for each feature
            map.
        """

        return super().compute(
            input_shape=input_shape,
            input_layer=input_layer,
            output_layers=output_layers
        )



from setuptools import setup

VERSION = '0.4.0'

setup(
    name='receptivefield',
    packages=['receptivefield'],
    version=VERSION,
    description='Gradient based Receptive field estimation library',
    author='fornax.ai',
    author_email='krzysztof.kolasinski@fornax.ai',
    url='https://github.com/fornaxai/receptivefield',
    download_url=f'https://github.com/fornaxai/receptivefield/archive/{VERSION}.tar.gz',
    keywords=['tensorflow', 'keras'],
    install_requires=[
        'pillow>=4.1.1',
        'matplotlib>=2.0.2',
        'numpy>=1.14.3',
    ],
    classifiers=[],
    include_package_data=True
)



from typing import NamedTuple, Any


class Size(NamedTuple):
    w: int
    h: int


class ImageShape(NamedTuple):
    w: int
    h: int
    c: int = 3


class GridPoint(NamedTuple):
    x: int
    y: int


class GridShape(NamedTuple):
    n: int
    w: int
    h: int
    c: int

    def replace(self, **kwargs: Any) -> "GridShape":
        return self._replace(**kwargs)


class ReceptiveFieldRect(NamedTuple):
    x: int
    y: int
    w: int
    h: int


class ReceptiveFieldDescription(NamedTuple):
    offset: GridPoint
    stride: GridPoint
    size: Size


class FeatureMapDescription(NamedTuple):
    """
    size: a feature map size
    rf: a ReceptiveFieldDescription
    """

    size: Size
    rf: ReceptiveFieldDescription


def to_rf_rect(point: GridPoint, size: Size) -> ReceptiveFieldRect:
    point = GridPoint(*point)
    size = Size(*size)
    return ReceptiveFieldRect(x=point.x, y=point.y, w=size.w, h=size.h)
